package pgupdate

import groovy.ui.SystemOutputInterceptor

import java.io.Console
import java.text.SimpleDateFormat


class SearchForPGInstances {
    def working_folder = System.getProperty("user.dir") + File.separator + "classes"
    def newline = System.getProperty("line.separator")

    //used to call MetadataConnection.groovy
    File MetadataConnectionClassFile = new File(working_folder, "MetadataConnection.groovy")
    Class MetadataConnectionClass = new GroovyClassLoader(getClass().getClassLoader()).parseClass(MetadataConnectionClassFile)
    GroovyObject metadataconnection = (GroovyObject) MetadataConnectionClass.newInstance()

    //used to call UpdatePostgres.groovy
    File UpdatePostgresClassFile = new File(working_folder, "UpdatePostgres.groovy")
    Class UpdatePostgresClass = new GroovyClassLoader(getClass().getClassLoader()).parseClass(UpdatePostgresClassFile)
    GroovyObject postgresupdater = (GroovyObject) UpdatePostgresClass.newInstance()

    //used to call UpdateLogger.groovy
    File UpdateLoggerClassFile = new File(working_folder, "UpdateLogger.groovy")
    Class UpdateLoggerClass = new GroovyClassLoader(getClass().getClassLoader()).parseClass(UpdateLoggerClassFile)
    GroovyObject updatelogger = (GroovyObject) UpdateLoggerClass.newInstance()

    //grab SASHOME environment variable
    def env = System.getenv()
    //replaces quotes as they are required for spaces on windows but will cause issues if left
    def SASHOME = env['SASHOME'].replaceAll('"', '')
    //Current list of existing DataServers to check for and update
    static def serverList = [[twelveByte:'WEBINFDSVRC', serverName:'WIP Data Server'],
        [twelveByte:'JOBMNTRDSVRC', serverName:'Job Monitoring Data Server'],
        [twelveByte:'FMDBSVRC', serverName:'Financial Management Data Server'],
        [twelveByte:'DMRMDTDSVRC', serverName:'Data Remediation Data Server'],
        [twelveByte:'CLADBSVRC', serverName:'Customer Link Analytics Data Server'],
        [twelveByte:'ABMDBSVRC', serverName:'Cost and Profitability Management Data Server'],
        [twelveByte:'AACMNDBSVRC', serverName:'Advanced Analytics Common Data Server'],
        [twelveByte:'DMDSVRC', serverName:'Data Management Data Server'],
        [twelveByte:'ORMONDSVRC', serverName:'Enterprise GRC Data Server'],
        [twelveByte:'CXADBSVRC', serverName:'Customer Experience Targeting Data Server'],
        [twelveByte:'WRTYDBSVRC', serverName:'Field Quality Analytics (formerly Warranty Analysis) Data Server'],
        [twelveByte:'DCMCOMDSVRC', serverName:'Decision Manager Common Data Server'],
        [twelveByte:'ENFCDBSVRC', serverName:'Energy Forecasting Data Server'],
        [twelveByte:'HPRISKDSVRC', serverName:'HP Risk Data Server'],
        [twelveByte:'MRMDSVRC', serverName:'Model Risk Management'],
        [twelveByte:'RGPDSVRC', serverName:'Risk Governance Platform'],
        [twelveByte:'DDCFDBSVRC', serverName:'Forecast Analyst Workbench'],
        [twelveByte:'INVTOPTDSVRC', serverName:'Inventory Optimization Workbench'],
        [twelveByte:'DMMONDSVRC', serverName:'Monitor and Dashboard Data Server']]

    /**
     * Used to call into the custom logger class
     * @return returns a new instance of the logger class
     */
    def log(){
        return UpdateLoggerClass.newInstance()
    }
    /**
     * Determines if the specified 12byte is installed on the system or not by checking the deployment registry.
     * @param twelveByte
     * @return
     */
    def Boolean isTwelvebyteInstalled(twelveByte){
        def registry_location = SASHOME + File.separator + "deploymntreg" + File.separator + "registry.xml"

        if((new File(registry_location).text).contains(twelveByte.toLowerCase())){
            return true
        }
        else{
            return false
        }
    }
    /**
     * Parses the deployment registry to find the corresponding SASHome installation location for the specified 12byte.
     * @param twelveByte
     * @return
     */
    def String getDatasrvHomeDir(twelveByte)
    {
        def registry_location = SASHOME + File.separator + "deploymntreg" + File.separator + "registry.xml"
        def Registry = new groovy.xml.XmlParser().parse(new File(registry_location))
        def location

        def installed_list = "${Registry.Key.Key.Key.Key.@name}".replaceAll(",", "").split()
        //println("Installed 12bytes: " + installed_list)

        Registry.Key.Key.Key.Key.findAll{ x ->
            x.@name.equalsIgnoreCase(twelveByte)//look for the 12byte block

        }.each{ entry ->
                 entry.children()[0].children().each{//once we find the block search for the sashome path
                    if(it.@name == "location")
                    {
                        //println it.@data
                        location = it.@data
                    }

                }
            }
        if (location != null){
            return location
        }
        else{
            return "The installation information for the following twelvebyte can not be found: " + twelveByte
        }
    }
    /**
     * Connects into MetadataConnection.groovy and grabs the dataserver information as a String.(ex: "datadir,user,pass")
     * @param twelveByte
     * @return
     */
    def String getDatasrvInfo(twelveByte)
    {
        def serverinfo = metadataconnection.getDataServerInfo(twelveByte)
        return serverinfo
    }

    /**
     * Gets the data directory from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvDataDir(serverInfo){
        return serverInfo.split(",")[0]
    }

    /**
     * Gets the configuration pg user from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvUser(serverInfo){
        return serverInfo.split(",")[1]
    }

    /**
     * Gets the pg password from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvPass(serverInfo){
        return serverInfo.split(",")[2]
    }

    /**
     * Gets the machine name stored in metadata from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvMachine(serverInfo){
        return serverInfo.split(",")[3]
    }

    /**
     * Gets the pg port from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvPort(serverInfo){
        return serverInfo.split(",")[4]
    }

    /**
     * Gets the PG_VERSION that is current in metadata from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvVersion(serverInfo){
        return serverInfo.split(",")[5]
    }

    /**
     * Gets the metadataid of the dataserver from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvMeta(serverInfo){
        return serverInfo.split(",")[6]
    }

    /**
     * Gets the metadataid of the dataserver from the passed in string
     * @param serverInfo
     * @return
     */
    def String getDatasrvName(serverInfo){
        return serverInfo.split(",")[7]
    }

    /**
     * Sets the PG_VERSION as a extended attribute in metadata
     * @param servermeta | metadataid of the dataserver
     * @param version
     * @return
     */
    def setPGVersionInMetadata(servermeta, version){
        metadataconnection.updatePgVersionExtension(servermeta, version)
    }

    /**
     * Creates the UIP_VERSION file as another indication that the upgrade has completed
     * @param datasrvdir_datadir | location of where the file will get created
     * @param version | version that will be reflected in the file
     * @return
     */
    def setPGVersionUIPonDisk(datasrvdir_datadir, version){
        log().info("Creating UIP_VERSION file in the following location to indicate upgrade completion: " + datasrvdir_datadir + File.separator + "UIP_VERSION")
        def file = new File(datasrvdir_datadir + File.separator + "UIP_VERSION")
        file.write(version)

        if (file.exists()){
            log().info("Successfully created UIP_VERSION file.")
        }
        else{
            log().error("Unable to successfully create UIP_VERSION file.")
        }
    }

    /**
     * Gets the version from the UIP_VERSION file on disk
     * @param datasrvdir_datadir | location of where the file lives
     * @return
     */
    def getPGVersionUIPonDisk(datasrvdir_datadir){
        def ver = 0
        def uip_verfile = new File(datasrvdir_datadir + File.separator + "UIP_VERSION")
        if(uip_verfile.exists()){
            ver = new BigDecimal(uip_verfile.text.trim())
        }
        return ver
    }

    /**
     * Gets the PG_VERSION that is current on disk in the data directory
     * @param datasrvdir_datadir | directory that contains the PG_VERSION file
     * @return
     */
    def String getDatasrvVersiononDisk(datasrvdir_datadir){
        def ver = 0
        def pg_verfile = new File(datasrvdir_datadir + File.separator + "PG_VERSION")
        if(pg_verfile.exists()){
            ver = new BigDecimal(pg_verfile.text.trim())
        }
        return ver
    }


    /**
     * Calls into UpdatePostgres.groovy to update a specific instance of a Postgres database.
     * @param twelvebyte
     * @param datasrvdir_home
     * @param datasrvdir_config
     * @param datasrvdir_datadir
     * @param datasrv_user
     * @param datasrv_pass
     * @return
     */
    def updateDBInstance(twelvebyte, datasrvdir_home, datasrvdir_config, datasrvdir_datadir, datasrv_user, datasrv_pass){
        postgresupdater.updateInstance(twelvebyte, datasrvdir_home, datasrvdir_config, datasrvdir_datadir, datasrv_user, datasrv_pass)
    }

    /**
     *Finds port by parsing the passed in startup script
     * @param datasrvdir_script
     * @return
     */
    def findPortFromScript(datasrvdir_script){
        return postgresupdater.findPortFromScript(datasrvdir_script)
    }

    /**
     *Finds data directory by parsing the passed in startup script
     * @param datasrvdir_script
     * @return
     */
    def findDataDirFromScript(datasrvdir_script){
        return postgresupdater.findDataDirFromScript(datasrvdir_script)
    }

    /**
     *Finds configuration directory by parsing the passed in startup script
     * @param datasrvdir_script
     * @return
     */
    def findConfigDirFromScript(datasrvdir_script){
        return postgresupdater.findConfigDirFromScript(datasrvdir_script)
    }

    /**
     *Finds the assumed product 12byte from the script file name
     * @param datasrvdir_script
     * @return
     */
    def findTwelveByteFromScript(datasrvdir_script){
        return postgresupdater.findTwelveByteFromScript(datasrvdir_script)
    }

    /**
     *Gets the version from the specified binaries
     * @param datasrvdir_bin
     * @return
     */
    def String getVersion(datasrvdir_home){
        return postgresupdater.getVersion(datasrvdir_home)
    }

    /**
     * Will determine to use shell or batch specification based on the current system OS
     * @return "sh" or "bat"
     */

    def static String getShellorBat(){
        String os = determineOS()

        if (os.equalsIgnoreCase("WIN")){
            return ".bat"
        }
        else{
            return ".sh"
        }

    }

    /**
     * Determines if the system is Unix or Windows
     * @return WIN or UNIX
     */

    def static String determineOS(){
        if (System.properties['os.name'].toLowerCase().contains('windows')) {
            return "WIN"
        } else {
            return "UNIX"
        }
    }

    /**
     * Sets up the logging options
     *
     */
    def initializeLogging(level){
        //initialize logfile
        String dateFormat = "yyyy.MM.dd_HH.mm.ss"
        SimpleDateFormat formatter = new SimpleDateFormat(dateFormat)
        String timestamp = formatter.format(new Date())

        if (System.properties['UPGRADE']){
            System.properties['LOGFILE'] = "upgrade_output_" + timestamp + ".log"
        }
        else{
            System.properties['LOGFILE'] = "analysis_output_" + timestamp + ".log"
        }

        File logfile = new File(System.properties['LOGFILE'])
        logfile.append("Postgres Update Log:" + newline)

        if(!level)
        {
            level = "INFO"
        }
        println("---------------")
        println("LOGLEVEL set to : " + level.toString().toUpperCase())
        println("---------------")
        if(level.toString().equalsIgnoreCase("TRACE")){
            System.properties['TRACELOGGING'] = true
            System.properties['DEBUGLOGGING'] = true
            System.properties['WARNLOGGING'] = true
        }
        else if(level.toString().equalsIgnoreCase("DEBUG")){
            System.properties['DEBUGLOGGING'] = true
            System.properties['WARNLOGGING'] = true
        }
        else if(level.toString().equalsIgnoreCase("WARN")){
            System.properties['WARNLOGGING'] = true
        }
        else{
            //null means log level info
        }
    }

    /**
     * Runs when code is in single instance mode
     *
     */
    def String runSingleInstance(twelvebyte, datasrvdir_home, datasrvdir_config, datasrvdir_datadir, datasrv_user, datasrv_pass, datasrv_version_uip, datasrv_version_disk, source_version, target_version){
        //def target_version = getVersion(datasrvdir_home) //want this as a number
        //def source_version = getVersion(datasrvdir_home + "_PREVIOUS") //want this as a number
        //if local and the path exists on the system
        //if(new File(datasrvdir_config).exists() && new File(datasrvdir_home).exists()){
        String result = ""
        if(new BigDecimal(datasrv_version_disk) < new BigDecimal(target_version)){
            if(System.properties['UPGRADE'] == true){
                //update the instance
                def update_rc = updateDBInstance(twelvebyte, datasrvdir_home, datasrvdir_config, datasrvdir_datadir, datasrv_user, datasrv_pass)

                //see if the update completed all the way though| set metadata value
                if(update_rc == 0){
                    log().info("Data Server was successfully updated: " + twelvebyte)
                    setPGVersionUIPonDisk(datasrvdir_datadir, target_version.toString())
                    result = "PASS"
                }
                else{
                    log().error("Data Server was NOT successfully updated: " + twelvebyte)
                    result = "FAIL"
                }

            }
            else{
                //analysis mode
            }
        }
        else{//version already set in metadata as 9.5
            log().info("Data Server already at specified version : " + target_version + " | " + twelvebyte)
            //log().error("The specified target dataserver version does not match the PostgreSQL binary version " + target_version)
        }
        return result
    }

    /**
     * Reads in a password from standard input while masking the input
     *@return password
     */
    def String readPassword (String prompt) {
        EraserThread et = new EraserThread(prompt)
        Thread mask = new Thread(et)
        mask.start()

        BufferedReader br = new BufferedReader(new InputStreamReader(System.in))
        String password = ""

              try {
                 password = br.readLine()
              } catch (IOException ioe) {
                ioe.printStackTrace();
              }
              // stop masking
              et.stopMasking()
              // return the password entered by the user
              return password

     }



    static void main(def args){
        def newline = System.getProperty("line.separator")
        SearchForPGInstances tc = new SearchForPGInstances()


        //handle commandline options
        def cli = new CliBuilder(usage: './run.sh --metaserver <host> --metaport <port> --metauser <user> --metapass <password> --dumploc <dumpdir> --upgrade --keepdump --deletebackup')
        def options
        cli.h(longOpt: 'help','Show usage information')
        cli.x(longOpt: 'upgrade', 'Run the upgrade', required: false)
        cli.kd(longOpt: 'keepdump', 'Keeps the dump file after execution', required: false)
        cli.db(longOpt: 'deletebackup', 'Deletes the backup data directory after execution', required: false)
        cli.si(longOpt: 'singleinstance', 'Specifies that user will be using the single instance mode and there for not connecting to metadata to retrieve information', required: false)
        cli.dl(args: 1, longOpt: 'dumploc', 'Specifies the location where you want the script to create dump files', required: false)
        cli.l(args: 1, longOpt: 'loglevel', 'set loglevel', required: false)
        options = cli.parse(args)

        if(options.si){
            println("Single Instance Mode(no metadata): requiring Additional Postgres parameters")
            cli.pgdatadir(args: 1, longOpt: 'pgdatadir', 'Postgres Data Directory', required: false)
            cli.pgconfig(args: 1, longOpt: 'pgconfig', 'Postgres Config Directory', required: false)
            cli.pgport(args: 1, longOpt: 'pgport', 'Postgres Data Server Port', required: false)
            cli.pguser(args: 1, longOpt: 'pguser', 'Postgres Admin User', required: true)
            cli.pgpass(args: 1, longOpt: 'pgpass', 'Postgres Password', required: false)
            cli.pgscript(args: 1, longOpt: 'pgscript', 'Path to Postgres Script', required: true)
        }
        else{
            println("Multiple Instance Mode(metadata required): requiring Additional Metadata parameters")
            cli.m(args: 1, longOpt: 'metaserver', 'Hostname for Metadata Server', required: true)
            cli.p(args: 1, longOpt: 'metaport', 'Port for Metadata Server', required: true)
            cli.u(args: 1, longOpt: 'metauser', 'Metadata Admin User', required: true)
            cli.pw(args: 1, longOpt: 'metapass', 'Metadata Admin Password', required: false)
        }

        if (options.arguments().isEmpty() || options.h){
            cli.usage()
            System.exit(1)
        }

        //parse additional parmeters
        options = cli.parse(args)

        //end if any required option is not specified
        if (options == null){
            System.exit(1)
        }

        //set the values passed in to be used later
        System.properties['METADATA_HOST'] = options.m
        System.properties['METADATA_PORT'] = options.p
        System.properties['METADATA_USER'] = options.u
        //set system-wide variables
        System.properties['UPGRADE'] = options.x
        System.properties['KEEPDUMP'] = options.kd
        System.properties['REMOVEDATABACKUP'] = options.db
        System.properties['SINGLEINSTANCEMODE'] = options.si
        System.properties['DUMPLOCATION'] = options.dl

        //initialize log
        tc.initializeLogging(options.l)

        //prompt for password if one was not provided
        if(options.pw == false && !options.si){
            try{
                System.properties['METADATA_PASS'] = tc.readPassword("> Enter your metadata admin password: ").toString()
            }catch(Exception ex){
                tc.log().error(ex + ": Problem getting password.")
                System.exit(-2)
            }
        }
        else{
            System.properties['METADATA_PASS'] = options.pw
        }

        tc.log().info("Starting execution.")
        tc.log().info("Processing registry for installed products.")

        //initialize output string for analysis mode
        def analysis_output = newline + newline + "9.4 SYSTEM POSTGRES ANALYSIS:" + newline + newline + newline
        def upgrade_output = newline + newline + "9.4 SYSTEM POSTGRES UPGRADE SUMMARY:" + newline + newline + newline

        //initialize variables
        def datasrv_counter_total = 0
        def datasrv_counter_local = 0
        def datasrv_counter_remote = 0
        def datasrv_counter_ready = 0
        def datasrv_counter_mismatch = 0
        def datasrvdir_home = ""
        def datasrvdir_config = ""
        def datasrvdir_datadir = ""
        def datasrv_user = ""
        def datasrv_pass = ""
        def datasrv_machine = ""
        def local_datasrv_counter_ready = 0
        def local_datasrv_machine = ""
        def remote_datasrv_machine = ""
        def local_msg = ""
        def remote_msg1 = ""
        def remote_msg2 = ""
        def datasrv_port = ""
        def datasrv_version_meta = 0
        def datasrv_version_disk = 0
        def datasrv_version_uip = 0
        def datasrv_meta = "N\\A"
        def datasrv_name = ""
        def twelvebyte = ""
        def islocal = true
        def source_version = 0
        def target_version = 0
        def source_version_flat = 0
        def target_version_flat = 0
        def backupdata = ""
        def reentry = false
        def upgrades_successful = 0
        def upgrades_failed = 0
        def dump_location = ""

        dump_location = System.properties['DUMPLOCATION']
        if (dump_location) {
            File file = new File(dump_location);
            if(!file.exists()){
                println("Specified dump location does not exist.")
                tc.log().error("Specified dump location " + dump_location + " does not exist.")
                System.exit(1)
            }

            if (!file.canWrite()) {
                println("Specified dump location is not writable.")
                tc.log().error("Specified dump location " + dump_location + " is not writable.")
                System.exit(1)
            }
        }

        //single instance mode
        if(System.properties['SINGLEINSTANCEMODE']){
            //twelvebyte (get 12byte from script path)
            datasrvdir_home = tc.getDatasrvHomeDir("WEBINFDSVR")//All data servers share this SASHOME location for the binaries

            //if the pgconfig, pgport, pgdata values are not specified via commandline, the code will attempt to get them from the pgscript. pgscript, pguser and pgpass are the only mandatory singleinstance parameters
            if(options.pgconfig == false){
                datasrvdir_config = tc.findConfigDirFromScript(options.pgscript)//get from script path
            }else{
                datasrvdir_config = options.pgconfig
            }

            if(options.pgport == false){
                datasrv_port = tc.findPortFromScript(options.pgscript)
            }else{
                datasrv_port = options.pgport
            }

            if(options.pgdata == false){
                datasrvdir_datadir = tc.findDataDirFromScript(options.pgscript)
            }else
            {
                datasrvdir_datadir = options.pgdata
            }

            twelvebyte = tc.findTwelveByteFromScript(options.pgscript)
            datasrv_user = options.pguser

            target_version = tc.getVersion(datasrvdir_home).toString().trim()
            source_version = tc.getVersion(datasrvdir_home + "_PREVIOUS").toString().trim()
            target_version_flat = target_version.toString().minus(".")
            source_version_flat = source_version.toString().minus(".")
            backupdata = datasrvdir_datadir + "_backup_" + source_version_flat

            tc.log().info("Binaries version of " + datasrvdir_home  + "_PREVIOUS is " + source_version)
            tc.log().info("Binaries version of " + datasrvdir_home  + " is " + target_version)

            //REENTRANT | check to make sure that datadirectory is there. If not then check for backupdir and get the next 2 values from there
            if(new File(datasrvdir_datadir).exists() == false){
                tc.log().info("The following data directory is not available: " + datasrvdir_datadir)
                if(new File(backupdata).exists()){
                    reentry = true
                    tc.log().info("Getting configured postgres instance information from a valid backup :" + backupdata)
                    datasrv_version_disk = tc.getDatasrvVersiononDisk(backupdata)
                    //datasrv_version_uip = tc.getPGVersionUIPonDisk(backupdata)
                    datasrv_version_uip = target_version
                }
            }
            else{
                tc.log().info("The following data directory is available: " + datasrvdir_datadir)
                datasrv_version_disk = tc.getDatasrvVersiononDisk(datasrvdir_datadir)
                //datasrv_version_uip = tc.getPGVersionUIPonDisk(datasrvdir_datadir)
                datasrv_version_uip = target_version
            }

            tc.log().debug("The following product 12byte has been detected in metadata: " + twelvebyte)
            tc.log().debug("Current dataserver version: " + datasrv_version_disk)
            tc.log().debug("Target dataserver version: " + datasrv_version_uip)

            datasrv_machine = InetAddress.localHost.hostName + "(local)"

            //check if a password was provided. If not prompt
            if(options.pgpass == false){
                try{
                    datasrv_pass = tc.readPassword("> Please Enter Your Postgres User Password: ").toString()
                }catch(Exception ex){
                    tc.log().error(ex + ": Problem getting password.")
                }
            }
            else{
                datasrv_pass = options.pgpass
            }

            tc.log().debug("SINGLE INSTANCE MODE!!!")

            String result = tc.runSingleInstance(twelvebyte, datasrvdir_home, datasrvdir_config, datasrvdir_datadir, datasrv_user, datasrv_pass, datasrv_version_uip, datasrv_version_disk, source_version, target_version)

            if(result == "PASS"){
                upgrades_successful++
            }
            else{
                upgrades_failed++
            }
            analysis_output = analysis_output.plus("DataServer: " + datasrv_name + "(" + twelvebyte +")" + newline)
            analysis_output = analysis_output.plus("=============================================" + newline)
            analysis_output = analysis_output.plus("DataServer Home Binary Location: " + datasrvdir_home + newline)
            analysis_output = analysis_output.plus("DataServer Data Location: " + datasrvdir_datadir + newline)
            analysis_output = analysis_output.plus("DataServer Configuration Location: " + datasrvdir_config + newline)
            analysis_output = analysis_output.plus("DataServer User: " + datasrv_user + newline)
            analysis_output = analysis_output.plus("DataServer Machine: " + datasrv_machine + newline)
            analysis_output = analysis_output.plus("DataServer Port: " + datasrv_port + newline)
            analysis_output = analysis_output.plus("DataServer Current Version: " + datasrv_version_disk + newline)

            if(new BigDecimal(datasrv_version_uip) != new BigDecimal(datasrv_version_disk) && new BigDecimal(datasrv_version_uip) != 0){
                analysis_output = analysis_output.plus("DataServer Target Version): " + datasrv_version_uip + newline)
                //nalysis_output = analysis_output.plus("ALERT! The version of the Postgres instance on disk does not match the version it should upon successful completion of the upgrade. This is could mean that there was a previous attempt at and upgrade that was unsuccessful." + newline)
                //datasrv_counter_mismatch++
            }
            if(datasrv_version_uip.toString() == "0" && datasrv_version_disk.toString() == target_version){
                analysis_output = analysis_output.plus("DataServer Version (Target version): N\\A" + newline)
                analysis_output = analysis_output.plus("ALERT! The version of the Postgres instance on disk appears to be at the correct version but its does not look like the previous upgrade completed successfully." + newline)
                //datasrv_counter_mismatch++
            }

            //increment local dataserver counter. Will always be 1 in this code path
            datasrv_counter_local++

            //determine if server is "ready" to upgrade

            //check if the script webinfdsvrc exists in datasrvdir
            File webinfdsvrc_file = new File(options.pgscript)
            if (!webinfdsvrc_file.exists()){
                tc.log().error("File " + webinfdsvrc_file + " does not exist. Run the command again after copying the file to your config directory.")
            }

            tc.log().debug("Property datasrv_version_uip is " + new BigDecimal(datasrv_version_uip))
            tc.log().debug("Property datasrv_version_disk is " + new BigDecimal(datasrv_version_disk))
            tc.log().debug("Property target_version is " + new BigDecimal(target_version))

            //condition 1
            if (new BigDecimal(datasrv_version_uip) != 0 && new BigDecimal(datasrv_version_disk) != 0){
                if (new BigDecimal(datasrv_version_disk) != new BigDecimal(source_version) && new BigDecimal(datasrv_version_disk) != new BigDecimal(target_version)) {
                    //datasrv_counter_ready++
                    //tc.log().info(twelvebyte + " is ready to upgrade. There is a previous target version that is < " + target_version)
                    tc.log().error(twelvebyte + " is ready to upgrade. However, the current version of binary SASHome/SASWebInfrastructurePlatformDataServer/9.4_PREVIOUS is " + source_version + ". The version must match the Data Server's current version " + datasrv_version_disk + ". Copy the version " + datasrv_version_disk + " binaries into the install directory and run the command again.")
                }
                else{
                    if (new BigDecimal(datasrv_version_uip) < new BigDecimal(target_version)){
                        datasrv_counter_ready++
                    }
                    else if (new BigDecimal(datasrv_version_uip) == new BigDecimal(target_version)) {
                            if (new BigDecimal(datasrv_version_disk) == new BigDecimal(target_version)) {
                                tc.log().info(twelvebyte + " seems already upgraded to the target version.")
                            }
                            else{
                                datasrv_counter_ready++
                            }
                    }
                    else{
                        tc.log().error(twelvebyte + " is higher than the target version.")
                    }
                }
            }
            //condition 2
            if(new BigDecimal(datasrv_version_disk) < new BigDecimal(target_version) && new BigDecimal(datasrv_version_uip) == 0){
                //datasrv_counter_ready++
                tc.log().info("No current version is found in the data directory and must be higher than 0. The current version is " + datasrv_version_disk)
            }

            //condition 3
            if(new BigDecimal(datasrv_version_disk) == new BigDecimal(target_version) && new BigDecimal(datasrv_version_uip) == 0 && new BigDecimal(datasrv_version_disk) != 0){
                //datasrv_counter_ready++
                tc.log().info(twelvebyte + " is ready to upgrade. There is not a previous target version. The current version is : " + datasrv_version_disk)
             }

            analysis_output = analysis_output.plus("=============================================" + newline)
            analysis_output = analysis_output.plus(newline)
        }
        //iterate through all of the 12bytes and check if they are installed on the system prior to attempting to get their information.
        else {
            //loop through the hardcoded list of 12bytes
            def datasrv_datadir_exist = false
            datasrv_counter_ready = 0
            datasrv_version_disk = 0
            datasrv_version_meta = 0
            serverList.each {
            twelvebyte = "${it['twelveByte']}"

            def datasrvdir_info = tc.getDatasrvInfo(twelvebyte)

            if(datasrvdir_info != "none"){
                tc.log().info("Processing: " + twelvebyte)
                datasrvdir_home = tc.getDatasrvHomeDir("WEBINFDSVR")//All data servers share this SASHOME location for the binaries
                datasrvdir_datadir = tc.getDatasrvDataDir(datasrvdir_info)
                datasrvdir_config = datasrvdir_datadir[0..-6]
                datasrv_user = tc.getDatasrvUser(datasrvdir_info)
                datasrv_pass = tc.getDatasrvPass(datasrvdir_info)
                datasrv_machine = tc.getDatasrvMachine(datasrvdir_info)
                datasrv_port = tc.getDatasrvPort(datasrvdir_info)
                datasrv_version_meta = tc.getDatasrvVersion(datasrvdir_info)
                datasrv_meta = tc.getDatasrvMeta(datasrvdir_info)
                datasrv_name = tc.getDatasrvName(datasrvdir_info)
                target_version = tc.getVersion(datasrvdir_home).toString().trim()//want this as a string
                source_version = tc.getVersion(datasrvdir_home + "_PREVIOUS").toString().trim()//want this as a string
                target_version_flat = target_version.minus(".")
                source_version_flat = source_version.minus(".")
                backupdata = datasrvdir_datadir + "_backup_" + source_version_flat
                islocal = false
                tc.log().info("Binaries version of " + datasrvdir_home  + "_PREVIOUS is " + source_version)
                tc.log().info("Binaries version of " + datasrvdir_home  + " is " + target_version)

                //REENTRANT | check to make sure that datadirectory is there. If not then check for backupdir and get the next 2 values from there
                datasrv_datadir_exist = new File(datasrvdir_datadir).exists()

                if(!datasrv_datadir_exist){
                    tc.log().warn("The following data directory is not available: " + datasrvdir_datadir)
                    if(new File(backupdata).exists()){
                        reentry = true
                        tc.log().warn("Getting configured postgres instance information from a valid backup :" + backupdata)
                        datasrv_version_disk = tc.getDatasrvVersiononDisk(backupdata)
                        //datasrv_version_uip = tc.getPGVersionUIPonDisk(backupdata)
                        datasrv_version_uip = target_version
                    }
                }
                else{
                    tc.log().info("The following data directory is available: " + datasrvdir_datadir)
                    datasrv_version_disk = tc.getDatasrvVersiononDisk(datasrvdir_datadir)
                    //datasrv_version_uip = tc.getPGVersionUIPonDisk(datasrvdir_datadir)
                    datasrv_version_uip = target_version
                }

                if(datasrv_datadir_exist){
                    tc.log().debug("The following product 12byte has been detected in metadata: " + twelvebyte)
                    tc.log().debug("Property datasrv_version_uip is " + new BigDecimal(datasrv_version_uip))
                    tc.log().debug("Property datasrv_version_disk is " + new BigDecimal(datasrv_version_disk))
                    tc.log().debug("Property target_version is " + new BigDecimal(target_version))
                }

                //server counters
                if (datasrv_datadir_exist) {
                    if(tc.isTwelvebyteInstalled(twelvebyte)){
                        datasrv_counter_local++
                        datasrv_machine = datasrv_machine + " (local)"
                        islocal = true
                    }
                }else{
                    datasrv_counter_remote++
                    remote_datasrv_machine = datasrv_machine
                    datasrv_machine = datasrv_machine + " (remote)"
                    islocal = false
                }

                //determine if server is "ready" to upgrade
                //check if the script webinfdsvrc exists in datasrvdir
                if (islocal && !datasrv_datadir_exist){
                    tc.log().error("File " + datasrvdir_datadir + " does not exist. Run the command again after copying the file to your config directory.")
                }

                //condition 1
                if (new BigDecimal(datasrv_version_uip) != 0 && new BigDecimal(datasrv_version_disk) != 0){
                    if (new BigDecimal(datasrv_version_disk) != new BigDecimal(source_version) && new BigDecimal(datasrv_version_disk) != new BigDecimal(target_version)) {
                        //tc.log().info(twelvebyte + " is ready to upgrade. There is a previous target version that is < " + target_version)
                        tc.log().error(twelvebyte + " is ready to upgrade. However, the current version of binary SASHome/SASWebInfrastructurePlatformDataServer/9.4_PREVIOUS is " + source_version + ". The version must match the Data Server's current version " + datasrv_version_disk + ". Copy the version " + datasrv_version_disk + " binaries into the install directory and run the command again.")
                    }
                    else{
                        if (new BigDecimal(datasrv_version_uip) < new BigDecimal(target_version)){
                            if (datasrv_datadir_exist){
                                datasrv_counter_ready++
                            }
                        }
                        else if (new BigDecimal(datasrv_version_uip) == new BigDecimal(target_version)) {
                                if (new BigDecimal(datasrv_version_disk) == new BigDecimal(target_version)) {
                                    tc.log().info(twelvebyte + " seems already upgraded to the target version.")
                                }
                                else{
                                    if (datasrv_datadir_exist){
                                        datasrv_counter_ready++
                                    }
                                }
                        }
                        else{
                            tc.log().error(twelvebyte + " is higher than the target version.")
                        }
                    }
                }

                //condition 2
                if(new BigDecimal(datasrv_version_disk) < new BigDecimal(target_version) && new BigDecimal(datasrv_version_uip) == 0 && new BigDecimal(datasrv_version_disk) != 0){
                    tc.log().info("No current version is found in the data directory and must be higher than 0. The current version is " + datasrv_version_disk)
                    //tc.log().info(twelvebyte + " is ready to upgrade. There is not a previous target version. The current version is : " + datasrv_version_disk)
                }

                //condition 3
                if(new BigDecimal(datasrv_version_disk) == new BigDecimal(target_version) && new BigDecimal(datasrv_version_uip) == 0 && new BigDecimal(datasrv_version_disk) != 0){
                    tc.log().info(twelvebyte + " is ready to upgrade. There is not a previous target version. The current version is : " + datasrv_version_disk)
                }

                //analysis values
                analysis_output = analysis_output.plus("=============================================" + newline)
                analysis_output = analysis_output.plus(datasrv_name + "(" + twelvebyte +")" + newline)
                analysis_output = analysis_output.plus("=============================================" + newline)
                analysis_output = analysis_output.plus("DataServer Home Binary Location: " + datasrvdir_home + newline)
                analysis_output = analysis_output.plus("DataServer Data Location: " + datasrvdir_datadir + newline)
                analysis_output = analysis_output.plus("DataServer Configuration Location: " + datasrvdir_config + newline)
                analysis_output = analysis_output.plus("DataServer User: " + datasrv_user + newline)
                analysis_output = analysis_output.plus("DataServer Machine: " + datasrv_machine + newline)
                analysis_output = analysis_output.plus("DataServer Port: " + datasrv_port + newline)

                if(reentry){
                    analysis_output = analysis_output.plus("DataServer Current Version: " + datasrv_version_disk + "(from existing backup: " + backupdata +  ")" +  newline)
                }
                else{
                    if(islocal){
                        analysis_output = analysis_output.plus("DataServer Current Version: " + datasrv_version_disk + newline)
                    }
                    else{
                        analysis_output = analysis_output.plus("DataServer Version (Metadata): " + datasrv_version_meta + newline)
                    }
                }

                if(datasrv_version_uip.toString() != datasrv_version_disk.toString() && datasrv_version_uip != 0){
                    analysis_output = analysis_output.plus("DataServer Target Version: " + datasrv_version_uip + newline)
                    //analysis_output = analysis_output.plus("ALERT! The current version of the Postgres instance does not match the version it should upon successful completion of the upgrade. This is could mean that there was a previous attempt at and upgrade that was unsuccessful." + newline)
                    //datasrv_counter_mismatch++
                }

                if(datasrv_version_uip.toString() == "0" && datasrv_version_disk.toString() == target_version){
                    analysis_output = analysis_output.plus("DataServer Target Version: N\\A" + newline)
                    analysis_output = analysis_output.plus("ALERT! The current version of the Postgres instance appears to be at the correct version but its does not look like the previous upgrade completed successfully." + newline)
                    //datasrv_counter_mismatch++
                }
                analysis_output = analysis_output.plus("=============================================" + newline)
                analysis_output = analysis_output.plus(newline)

                //if the UIP_VERSION does not exist or is at a lower version than the target
                if (new BigDecimal(datasrv_version_disk) < new BigDecimal(target_version)){
                    //check if the --upgrade option set
                    if (islocal && System.properties['UPGRADE'] == true){
                        //update the instance
                        def update_rc = tc.updateDBInstance(twelvebyte, datasrvdir_home, datasrvdir_config, datasrvdir_datadir, datasrv_user, datasrv_pass)

                        //see if the update completed all the way though| set metadata value
                        if(update_rc == 0){
                            tc.log().info("Data Server was successfully updated: " + twelvebyte)
                            tc.setPGVersionUIPonDisk(datasrvdir_datadir, target_version)
                            //only set the version if it is not already set
                            if(datasrv_version_meta != target_version){
                                tc.setPGVersionInMetadata(twelvebyte, target_version)
                            }

                            upgrade_output = upgrade_output.plus("=============================================" + newline)
                            upgrade_output = upgrade_output.plus(datasrv_name + "(" + twelvebyte +") upgrade SUCCESSFUL" + newline)
                            upgrade_output = upgrade_output.plus("=============================================" + newline)
                            upgrades_successful++
                        }
                        else{
                            tc.log().error("Data Server was NOT successfully updated: " + twelvebyte)
                            upgrade_output = upgrade_output.plus("=============================================" + newline)
                            upgrade_output = upgrade_output.plus(datasrv_name + "(" + twelvebyte +") upgrade NOT SUCCESSFUL" + newline)
                            upgrade_output = upgrade_output.plus("=============================================" + newline)
                            upgrades_failed++
                        }
                    }
                    else{
                        //analysis mode
                    }
                }
                else{//version already set in metadata as 9.5
                    tc.log().info("Data Server already at specified version: " + target_version + " | " + datasrv_name + "(" + twelvebyte + ")")
                    //tc.log().error("The specified target dataserver version does not match the PostgreSQL binary version " + target_version + " | " + datasrv_name + "(" + twelvebyte +")")
                }
            }
            //12byte installed but not configured || or could not get metadata back
            else{
                //println("The following product 12byte was not configured: " + twelvebyte)
            }

            }//end 12byte loop
        }

        datasrv_counter_total = datasrv_counter_remote + datasrv_counter_local

        //if ANALYSIS MODE print the ANALYSIS OUTPUT
        if(System.properties['UPGRADE'] != true){
            analysis_output = analysis_output.plus("Total Number of Data Servers: " + datasrv_counter_total + newline)
            analysis_output = analysis_output.plus("Total Number of Local Data Servers: " + datasrv_counter_local + newline)
            if (datasrv_counter_remote> 0) {
                analysis_output = analysis_output.plus("Total Number of Remote Data Servers: " + datasrv_counter_remote + newline)
            }

            if(datasrv_counter_mismatch > 0){
                analysis_output = analysis_output.plus("Total Number of Data Servers with Conflicting Versions: " + datasrv_counter_mismatch + newline)
            }

            local_datasrv_counter_ready = datasrv_counter_ready
            
            analysis_output = analysis_output.plus("Total Number of Data Servers Ready to Upgrade on this local server: " + local_datasrv_counter_ready + newline)

            if(datasrv_counter_ready > 0){
                local_msg   = "To upgrade the data servers on this local server, run the UpgradePostgres script with '--upgrade' parameter."
                if(datasrv_counter_remote > 0) {
                   remote_msg1 = "The UpgradePostgres script has detected that one or more Postgres Data Servers are configured on " + remote_datasrv_machine + "."
                   remote_msg2 = "Once you have completed the upgrade for the data servers on this local server, you will need to run the Upgrade Postgres steps on that server as well."
                   analysis_output = analysis_output.plus(newline + remote_msg1 + newline + remote_msg2 + newline + local_msg + newline)
                }
                else{
                    analysis_output = analysis_output.plus(newline + local_msg + newline)
                }
            }
            else{
                analysis_output = analysis_output.plus(newline + "All local discovered dataservers are at the correct version." + newline)
            }
            tc.log().info(analysis_output)

            //new File("analysis.txt").write(analysis_output)
        }
        else
        {
            if(datasrv_counter_ready > 0){
                //print upgrade summary
                upgrade_output = upgrade_output.plus("Successful Upgrades: " + upgrades_successful + newline)
                upgrade_output = upgrade_output.plus("Unsuccessful Upgrades: " + upgrades_failed + newline)
                tc.log().info(upgrade_output)
            }
            else{
                tc.log().info("All local discovered dataservers are at the correct version.")
            }
        }

        tc.log().info("Ending execution")
        System.exit(0)

    }

    class EraserThread implements Runnable {
        private boolean stop;

        /**
         *@param The prompt displayed to the user
         */
        public EraserThread(String prompt) {
            System.out.print(prompt);
        }

        /**
         * Begin masking...display asterisks (*)
         */
        public void run () {
           stop = true;
           while (stop) {
              System.out.print("\010*");
          try {
             Thread.currentThread().sleep(1);
              } catch(InterruptedException ie) {
                 ie.printStackTrace();
              }
           }
        }

        /**
         * Instruct the thread to stop masking
         */
        public void stopMasking() {
           this.stop = false;
        }
     }
}
